@@ -48,7 +48,6 @@ impl AttributedForeignPredicateDefinition {
 
 impl Parse for AttributedForeignPredicateDefinition {
     fn parse(input: ParseStream) -> Result<Self> {
-        let visibility = input.parse()?;
         let attrs: Vec<Attribute> = input.call(Attribute::parse_outer)?;
         let mut doc = None;
         let mut predicate_name = None;
@@ -63,6 +62,7 @@ impl Parse for AttributedForeignPredicateDefinition {
             }
         }
 
+        let visibility = input.parse()?;
         let predicate = input.parse()?;
 
         Ok(Self {
@@ -208,7 +208,7 @@ impl ForeignPredicateDefinitionImpl for SemidetForeignPredicateDefinition {
                     arity: std::os::raw::c_int,
                     _control: *mut std::os::raw::c_void
                 ) -> isize {
-                    let result = std::panic::catch_unwind(|| {
+                    let result = #crt::context::prolog_catch_unwind(|| {
                         if #known_arity as usize != arity as usize {
                             // TODO actually throw an error
                             return Err(#crt::result::PrologError::Failure);
@@ -229,37 +229,12 @@ impl ForeignPredicateDefinitionImpl for SemidetForeignPredicateDefinition {
                                          #(#term_args),*)
                     });
 
-                    // creating a new context here cause context is not allowed to cross the catch boundary.
-                    let context = #crt::context::unmanaged_engine_context();
                     match result {
                         Ok(result) => match result {
                             Ok(()) => 1,
                             Err(_) => 0,
                         },
-                        Err(panic) => {
-                            let panic_term = context.new_term_ref();
-                            let error_term = term!{context: error(rust_error(panic(#&panic_term)), _)};
-
-                            match panic.downcast_ref::<&str>() {
-                                Some(panic_msg) => {
-                                    panic_term.unify(panic_msg).unwrap();
-                                }
-                                None => {
-                                    match panic.downcast_ref::<String>() {
-                                        Some(panic_msg) => {
-                                            panic_term.unify(panic_msg.as_str()).unwrap();
-                                        },
-                                        None => {
-                                            panic_term.unify("unknown panic type").unwrap();
-                                        }
-                                    }
-                                }
-                            }
-
-                            context.raise_exception::<()>(&error_term).unwrap_err();
-
-                            0
-                        }
+                        Err(_) => 0
                     }
                 }
             },
@@ -293,7 +268,7 @@ impl ForeignPredicateDefinitionImpl for SemidetForeignPredicateDefinition {
             #visibility fn #registration_name() -> bool {
                 // unsafe justification: register_foreign_in_module is unsafe due to the possibility that someone registers a function that is not actually expecting to handle a prolog call, and is not set up right for it. But we're generating this code ourselves, so we should have taken care of all preconditions.
                 unsafe {
-                    #crt::engine::register_foreign_in_module(
+                    #crt::init::register_foreign_in_module(
                         #module_lit,
                         #name_lit,
                         std::convert::TryInto::<u16>::try_into(#arity).expect("arity does not fit in an u16"),
@@ -393,7 +368,7 @@ impl ForeignPredicateDefinitionImpl for NondetForeignPredicateDefinition {
         let data_name = &self.data_name;
         let data_type = &self.data_type;
         quote! {
-            fn #definition_setup_name<'a, C: #crt::context::QueryableContextType>(#[allow(unused)] #context_arg: &'a #crt::context::Context<'a, C>, #(#[allow(unused)] #term_args : &#crt::term::Term<'a>),*) -> #crt::result::PrologResult<#data_type> {
+            fn #definition_setup_name<'a, C: #crt::context::QueryableContextType>(#[allow(unused)] #context_arg: &'a #crt::context::Context<'a, C>, #(#[allow(unused)] #term_args : &#crt::term::Term<'a>),*) -> #crt::result::PrologResult<Option<#data_type>> {
                 #setup_code
             }
 
@@ -433,7 +408,7 @@ impl ForeignPredicateDefinitionImpl for NondetForeignPredicateDefinition {
                     // first, assert that the data type is send and unpin
                     #trampoline_type_check::<#data_type>();
 
-                    let result = std::panic::catch_unwind(|| {
+                    let result = #crt::context::prolog_catch_unwind(|| {
                         if #known_arity as usize != arity as usize {
                             // TODO actually throw an error
                             return Err(#crt::result::PrologError::Failure);
@@ -455,8 +430,15 @@ impl ForeignPredicateDefinitionImpl for NondetForeignPredicateDefinition {
                                 // this is the first call
                                 let result = #definition_setup_name(&context,
                                                                #(#term_args),*)?;
+                                if let Some(result) = result {
+                                    // proceed
+                                    data = Box::new(result);
+                                }
+                                else {
+                                    // we're done
+                                    return Ok(None);
+                                }
 
-                                data = Box::new(result);
                             },
                             2 => {
                                 // this is a subsequent call - there should already be data.
@@ -493,38 +475,13 @@ impl ForeignPredicateDefinitionImpl for NondetForeignPredicateDefinition {
                         Ok(retry)
                     });
 
-                    // creating a new context here cause context is not allowed to cross the catch boundary.
-                    let context = #crt::context::unmanaged_engine_context();
                     match result {
                         Ok(result) => match result {
                             Ok(None) => 1,
                             Ok(Some(r)) => r as isize,
                             Err(_) => 0,
                         },
-                        Err(panic) => {
-                            let panic_term = context.new_term_ref();
-                            let error_term = term!{context: error(rust_error(panic(#&panic_term)), _)};
-
-                            match panic.downcast_ref::<&str>() {
-                                Some(panic_msg) => {
-                                    panic_term.unify(panic_msg).unwrap();
-                                }
-                                None => {
-                                    match panic.downcast_ref::<String>() {
-                                        Some(panic_msg) => {
-                                            panic_term.unify(panic_msg.as_str()).unwrap();
-                                        },
-                                        None => {
-                                            panic_term.unify("unknown panic type").unwrap();
-                                        }
-                                    }
-                                }
-                            }
-
-                            context.raise_exception::<()>(&error_term).unwrap_err();
-
-                            0
-                        }
+                        Err(_) => 0
                     }
                 }
             },
@@ -558,7 +515,7 @@ impl ForeignPredicateDefinitionImpl for NondetForeignPredicateDefinition {
             #visibility fn #registration_name() -> bool {
                 // unsafe justification: register_foreign_in_module is unsafe due to the possibility that someone registers a function that is not actually expecting to handle a prolog call, and is not set up right for it. But we're generating this code ourselves, so we should have taken care of all preconditions.
                 unsafe {
-                    #crt::engine::register_foreign_in_module(
+                    #crt::init::register_foreign_in_module(
                         #module_lit,
                         #name_lit,
                         std::convert::TryInto::<u16>::try_into(#arity).expect("arity does not fit in an u16"),
